/* file name  : parser/events/metaevents/MetaPropagationManager.java
 * authors    : Guillaume Arnaud
 * created    : mar 23 mar 2004 14:37:47 CET
 * copyright  : 
 *
 * modifications:
 *
 */

package parser.propagation;

import parser.PropagationHandler;
import parser.propagation.MetaPropagationSolverInterface;
import request.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;

import fr.inria.oadymppac.parsing.ContentHandler;
import fr.inria.oadymppac.parsing.event.EventListenerImpl;
import fr.inria.oadymppac.parsing.gentra4cp.event.ReduceEvent;
import fr.inria.oadymppac.parsing.gentra4cp.event.ReduceEventListener;
import fr.inria.oadymppac.parsing.gentra4cp.event.AwakeEvent;
import fr.inria.oadymppac.parsing.gentra4cp.event.AwakeEventListener;
import fr.inria.oadymppac.parsing.gentra4cp.event.SolutionEvent;
import fr.inria.oadymppac.parsing.gentra4cp.event.SolutionEventListener;
import fr.inria.oadymppac.parsing.gentra4cp.event.FailureEvent;
import fr.inria.oadymppac.parsing.gentra4cp.event.FailureEventListener;
import fr.inria.oadymppac.parsing.gentra4cp.event.ChoicePointEventListener;
import fr.inria.oadymppac.parsing.gentra4cp.event.ChoicePointEvent;
import fr.inria.oadymppac.parsing.gentra4cp.event.NewConstraintEvent;
import fr.inria.oadymppac.parsing.gentra4cp.event.NewConstraintEventListener;
import fr.inria.oadymppac.parsing.gentra4cp.event.attributes.EventAttributes;

/**
 * MetaPropagationManager is a class that collects the 
 * propagation events, inherited from {@link PropagationEvent},  between two solver events
 * inherited from {@link EventAttributes}. ChoicePointEvent, SolutionEvent and FailureEvent
 * delimit each propagation phase. NewConstraintEvent allows to record constraint
 * informations which are necessary for particular filtered constraint queries. Indeed 
 * the user can make a request on a particular propagation event impliying
 * specified constraints thanks to {@link PropagationRequest}. ReduceEvent
 * and AwakeEvent are the listened PropagationEvents. During a propagation phase 
 * a counter for each request and a counter of all queried propagation events are incremented,
 * queried constraints are record if they are reduced or awaked.
 *  
 * @author Guillaume Arnaud
 * @version 1.0
 */

public class  MetaPropagationManager extends EventListenerImpl
	implements ReduceEventListener, AwakeEventListener, SolutionEventListener, FailureEventListener,  ChoicePointEventListener, NewConstraintEventListener {	
	/**
	 * The solver stores the event listeners. 
     * @see PropagationHandler
	 */
	private PropagationHandler solver;

	/**
	 * Requests which specify the propagation events and constraint which will be analyzed. 
	 */
	private PropagationRequest requests;

	/**
	 * Stores a list of solvers which listen meta propagation event.
     * Notice: In general case there are just one solver.
	 */
	private Vector solverList;

	/**
     * Total number of propagations, specified in the request, between two control events.
	 */
	private int nbTotalPropagation=0;
	
	/** 
	 * Total number of reduces between two control events.
	 */
	private int nbRed=0;
	
	/** 
	 * Total number of awake between two control events.
	 */
	private int nbAwake=0;
    
	/** 
	 * Vector that stores all new constraints occured.
	 */
	private Vector constraintNames;

	/** 
	 * Hashtable which associates, for each awake event, a filtered constraint and a
     * counter of its occurence.
	 */
	private Hashtable awakeTable;
    
	/** 
	 * Hashtable which associates, for each reduce event, a filtered constraint and a
     * counter of its occurence.
	 */    
	private Hashtable reduceTable;
    

	/**
	 * Correspondance table between the external representation and the identifier.
	 */
	private Hashtable correspondanceRep;
	
	/** 
	 * Constructor.
	 */
	public MetaPropagationManager(){ }
    
    
	/** 
	 * Initialized the solver and the propagation request. Only the propagation events
     * present into request will be listened.
	 * 
	 * @param solver the PropagationHandler which generates events
	 * @param pr the PropagationRequest which contains requests
	 */    
	public MetaPropagationManager(PropagationHandler solver, PropagationRequest pr){
		this.solver=solver;
		this.requests=pr;

        try {
        solver.addNewConstraintListener(this,false);
		if(requests.hasRequest("reduce")) {
            solver.addReduceListener(this);
            reduceTable=new Hashtable();
        }
		if(requests.hasRequest("awake")) {
            solver.addAwakeListener(this);
            awakeTable=new Hashtable();
        }

		solver.addChoicePointListener(this);
		solver.addSolutionListener(this);
		solver.addFailureListener(this);
        } catch (ClassNotFoundException e) {
            ContentHandler.logger.error("Impossible to find class of Event", e);
        } catch (InstantiationException e) {
            ContentHandler.logger.error(
                    "Impossible to generate a new instance of Event", e);
        } catch (IllegalAccessException e) {
            ContentHandler.logger.error(
                    "Illegal method call to generate a new instance of Event",
                    e);
        } catch (Exception e) {
            ContentHandler.logger.error("Undetermined Exception", e);
        }
        
	}
	
    /** 
     * Initialized the solver.
     * 
     * @param solver  the PropagationHandler which generates events
     */
    public MetaPropagationManager(PropagationHandler solver){
        try {
            this.solver=solver;
    		solver.addChoicePointListener(this);
    		solver.addSolutionListener(this);
	    	solver.addFailureListener(this);}
         catch (ClassNotFoundException e) {
            ContentHandler.logger.error("Impossible to find class of Event", e);
        } catch (InstantiationException e) {
            ContentHandler.logger.error(
                    "Impossible to generate a new instance of Event", e);
        } catch (IllegalAccessException e) {
            ContentHandler.logger.error(
                    "Illegal method call to generate a new instance of Event",
                    e);
        } catch (Exception e) {
            ContentHandler.logger.error("Undetermined Exception", e);
        }
            
    }

    /** 
     * Initializes the propagation data.
     */
    private void initData(){
        nbRed=0;
        nbAwake=0;
		nbTotalPropagation=0;

        if(awakeTable!=null) awakeTable.clear();
        if(reduceTable!=null) reduceTable.clear();
        if(constraintNames!=null) constraintNames.clear();
    }

	/**
	 * Adds a propagation listener.
     *
	 * @param l MetaPropagationListener
	 */
	public void addMetaPropagationListener(MetaPropagationListener l) {
        if(solverList==null) this.solverList=new Vector();
		this.solverList.add(l);
	}
    
	/** 
	 * Removes a listener of a list list.
	 * 
	 * @param listener 
	 * @param list 
	 * @return 
	 */
    public void removeMetaPropagationListener(MetaPropagationListener l) {
    	for (int i = 0; i < solverList.size(); i++) {
    	    if (solverList.elementAt(i) == l) { solverList.removeElementAt(i); }
    	}
    }

    /** 
     * Does have listener?
     * 
     * @return boolean
     */
    public boolean hasMetaPropagationListener() {return solverList!=null;   }
    
    
    /** 
     * Gets the meta propagation listener at index.
     * 
     * @param index 
     * @return MetaPropagationListener
     */
    public MetaPropagationListener getMetaPropagationListener(int index) {
	    return (MetaPropagationListener)solverList.elementAt(index);
    }
    
	/**
	 * When a SolutionEvent occurs a EventAttributes is 
	 * raised with the current propagation data.  
	 * @param e SolutionEvent
	 */
	public void eventProcess(SolutionEvent e){        
		this.fireMetaPropagationListener(MetaPropagationEvent.SOLUTION,e);
        this.initData();
	}
	
	/*
	 * When a FailureEvent occurs a EventAttributes is 
	 * raised with the current propagation data. 
	 * @param e FailureEvent
	 */
	public void eventProcess(FailureEvent e){
		this.fireMetaPropagationListener(MetaPropagationEvent.REJECT,e);
        this.initData();
	}
    
	/**
	 * When a ChoicePointEvent occurs a EventAttributes is 
	 * raised with the current propagation data. 
	 * @param e ChoicePointEvent
	 */
	public void eventProcess(ChoicePointEvent e){
		fireMetaPropagationListener(MetaPropagationEvent.CHOICEPOINT,e);
        this.initData();
	}


	/**
	 * When a ReduceEvent occurs nbRed is incremented. The external representation of constraints 
     * queried by the requests are parsed, by convention, until the first opening bracket '('.
	 * @param e ReduceEvent
	 */
	public void eventProcess(ReduceEvent e){
        //System.out.println("[MetaPropagationManager] Reduce chrono="+e.getChrono());
		nbTotalPropagation++;
		if(requests==null || !requests.hasRequest("reduce")) return;
		nbRed++;
		if(!requests.isFilter("reduce")||e.getConstraint()==null || this.correspondanceRep==null) return;
		else {
			String cIdent=e.getConstraint();
			String extRep=(String)correspondanceRep.get(cIdent);
			int index;
			Vector constraintFilter=requests.getConstraintFilters("reduce");
			Enumeration cst;			
			
			if(constraintFilter==null) return;
			index=extRep.indexOf("(")-1;
			cst=constraintFilter.elements();
			while(cst.hasMoreElements()) {
				String constraintCurr=((String)cst.nextElement());
				if(extRep.regionMatches(1,constraintCurr,1,index)){
					if(reduceTable.containsKey(constraintCurr)){
						int cpt=((Integer)reduceTable.get(constraintCurr)).intValue();
						cpt++;
						reduceTable.put(constraintCurr,new Integer(cpt));
					}
					else{
						reduceTable.put(constraintCurr,new Integer(1));}
				}
			}
		}

	}

	
	/**
	 * When a AwakeEvent occurs nbAwake is incremented. The external representation of constraints 
     * queried by the requests are parsed, by convention, until the first opening bracket '('.
	 * @param e AwakeEvent
	 */
	public void eventProcess(AwakeEvent e){
        //System.out.println("[MetaPropagationManager] Awake chrono="+e.getChrono());
		nbTotalPropagation++;
		if(requests==null || !requests.hasRequest("awake")) return;
		nbAwake++;		
		if(!requests.isFilter("awake")||e.getConstraint()==null || this.correspondanceRep==null) return;
		else
		{
			String cIdent=e.getConstraint();
			String extRep=(String)correspondanceRep.get(cIdent);
			int index;
			Vector constraintFilter=requests.getConstraintFilters("awake");
			Enumeration cst;			
			
			if(constraintFilter==null) return;
			if(awakeTable==null) awakeTable=new Hashtable();
			index=extRep.indexOf("(")-1;
			cst=constraintFilter.elements();
			while(cst.hasMoreElements()) {
				String constraintCurr=((String)cst.nextElement());
				if(extRep.regionMatches(1,constraintCurr,1,index)){
					if(awakeTable.containsKey(constraintCurr)){
						int cpt=((Integer)awakeTable.get(constraintCurr)).intValue();
						cpt++;
						awakeTable.put(constraintCurr,new Integer(cpt));
					}
					else{
						awakeTable.put(constraintCurr,new Integer(1));}
				}
			}
		}

	}
    
	/**
     * Stores the external representation of the new constraint and associated it
     * with its identifier. 
	 * @param e NewConstraintEvent
	 */
	public void  eventProcess(NewConstraintEvent e){
        String name=e.getName();
        
        if(name.equals("- NONAMED CONSTRAINT -"))
                name=e.getInternal();
        else return;        
        if(correspondanceRep==null) correspondanceRep=new Hashtable();
		if(e.getIdentifier()!=null) correspondanceRep.put(e.getIdentifier(),name);
		if(constraintNames==null) constraintNames=new Vector();
        constraintNames.addElement(name);

	}

	
	/**
	 * Fires MetaPropagationEvent into the registered listeners.
	 * @param typeEvent 
	 * @param solevent 
	 */
	public void fireMetaPropagationListener(int typeEvent, EventAttributes solevent) {
		if (solverList==null || solverList.size() == 0) {
		    return;
		}
		MetaPropagationEvent e = new MetaPropagationEvent(solevent,typeEvent,nbTotalPropagation, constraintNames);
		if(requests!=null && requests.getRequests()!= null) {
			if (requests.getRequests().containsKey("awake")){
				e.addCounter("awake",nbAwake,awakeTable); }			
			if (requests.getRequests().containsKey("reduce")){
				e.addCounter("reduce",nbRed,reduceTable); }
		}

		for (int i = 0; i < solverList.size(); i++) {
			MetaPropagationSolverInterface l =
				(MetaPropagationSolverInterface)solverList.elementAt(i);
			l.notifyMetaPropagationListener(e);
		}
	}
}

