/* file name  : request/PropagationRequest.java
 * authors    : Guillaume Arnaud
 * created    : mar 23 mar 2004 16:59:06 CET
 * copyright  : 
 *
 * modifications:
 *
 */

package request;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * Stores requests for propagation. Propagation events are keys of an hashtable
 * {@link #requests} and associated to an inner class {@link #PropagationOneRequest}.
 * This last class contains a vector of constraint names that we want to filter.
 * 
 * @author Guillaume Arnaud
 * @version 1.0
 */
public class PropagationRequest extends Object {
	
	/** 
	 * Hashtable of requests. An event type is associated to a {@link
     * #PropagationOneRequest}.
	 */
	private Hashtable requests=null;
	
	/** 
	 * Vector of different event types requested.
	 */
	private Vector typeList=null;
	
	/**
	 */
	public PropagationRequest(){
    }

	/**
	 * Adds a request. 
     *
	 * @param t name of a propagation event (inherited from {@link PropagationEvent})
	 * @param f boolean specifying if this event is filtered
	 * @param c Vector containing filtered constraint name (null if there are none)
	 */
	public void addRequest(String t, boolean f, Vector c){
		PropagationOneRequest newRequest=new PropagationOneRequest(t,f);
		
		if(requests!=null && requests.containsKey(t)) {
			System.out.println("Error PropagationRequest.addRequest(String,boolean,Vector) : Existing type of request");
			return;
		}
		
		if(requests==null) {
			requests=new Hashtable();
			typeList=new Vector();
		}
		if(f&&c!=null){
			Enumeration e=c.elements();
			while(e.hasMoreElements()){
				newRequest.addConstraintFilter((String)e.nextElement());}
			requests.put(t,newRequest);
			typeList.addElement(t);
		}
		else {
			requests.put(t,newRequest);
		}
	}
	
	/**
	 * Adds a filtered constraint to the event. If the event doesn't exist before 
     * the addRequest method is invoked.
     *
	 * @param t name of a propagation event (inherited from {@link PropagationEvent})
	 * @param c name of a filtered constraint
	 */
	public void addConstraintFilter(String t, String c){
		if(requests!=null && requests.containsKey(t)) {
			PropagationOneRequest por=(PropagationOneRequest)requests.get(t);
			por.addConstraintFilter(c);
            return;
		}
        Vector tabfilters=new Vector();
        tabfilters.add(c);
        this.addRequest(t,true,tabfilters);
        
	}

	/**
	 * Has the request t?.
     * 
	 * @param t name of the queried event
	 * @return  boolean response
	 */
	public boolean hasRequest(String t){
		if(requests!=null && requests.containsKey(t)) return true;	
		else return false;		
	}
    
	/** 
	 * Has request?
	 * 
	 * @return boolean response
	 */
	public boolean hasRequest(){
		return requests!=null;		
	}
	
	/**
	 * Gets requests. 
     * 
	 * @return Hashtable of requests.
	 */
	public Hashtable getRequests() { return requests;}
    
	/**
	 * t is a filtered request?
     * 
	 * @param t name of the propagation event
	 * @return boolean response
	 */
	public boolean isFilter(String t){
		if(requests==null || !requests.containsKey(t)) return false;
		else return ((PropagationOneRequest)requests.get(t)).filter;
	}
	
	/**
	 * Gets type list.
     * 
	 * @return Vector of type.
	 */
	public Vector getTypeList(){ return typeList;}

	/**
	 * Gets filtered constraint.
	 * @param t name of the propagation event
	 * @return Vector of filtered constraint names
	 */
	public Vector getConstraintFilters(String t){
		if(requests.containsKey(t))
			return ((PropagationOneRequest)requests.get(t)).getConstraintFilter();
		else return null;
	}
	
	/** 
	 * Request associated to one propagation event.
	 * 
	 * @author Guillaume Arnaud
	 * @version 1.0
	 */
	private class PropagationOneRequest extends Object {
		
		/** 
		 * Are there filters?
		 */
		boolean filter=false; 
		
		/** 
		 * Event type of this request (reduce, awake)
		 */
		String type="";
		
		/** 
		 * Vector of filtered constraint name.
		 */
		Vector constraintFilter=null;

		/**
		 * Constructor.
		 */
		public PropagationOneRequest(){
		}

		/** 
		 * Builds a request of event type t and containing filtered constraint c.
		 * 
		 * @param t event type name.
		 * @param c filtered constraint name.
		 */
		public PropagationOneRequest(String t, String c) {
			filter=true;
			type=t;
			constraintFilter=new Vector();
			constraintFilter.addElement(c);
		}
		
		/** 
		 * Builds a request of event type t with the indication of filter.
		 * 
		 * @param t event type name. 
		 * @param f boolean, true if request will contain filters, false else.
		 */
		public PropagationOneRequest(String t, boolean f) {
			filter=f;
			type=t;
			if(f) constraintFilter=new Vector();
		}

		/**
		 * Adds a constraint filter.
         * 
		 * @param c constraint filter.
		 */
		public void addConstraintFilter(String c) {
			if(constraintFilter==null) {
				constraintFilter=new Vector();
                filter=true;
            }
			constraintFilter.addElement(c);
		}

		/** 
		 * Gets the vector of constraint filters.
		 * 
		 * @return vector of constraint filters.
		 */
		public Vector getConstraintFilter() {return constraintFilter;}
	}

}
