// A Graphical User Interface for Constraint Logic Programming
// http://contraintes.inria.fr/~fages/CLPGUI
// Copyright (C) 2001, 2002, INRIA Francois Fages
// License LGPL-2.1

// Version 2.1.3
// file CLPGUI.java to use with the GNU-Prolog library clpgui.pl

// 24 Nov 2001: Version 0
// - single window interface
// - connection to GNU-Prolog
// - visualization of Finite Domains
// - interactions with the solver
// 11 Jan 2002: Version 1
// - multiwindow interface
// - visualization of the search tree
// - inner classes for viewers
// 23 August 2002: Version 2.1.1
// - recomputations
// 27 September 2002:  Version 2.1.2
// - 2D domains viewer and communication of domains with GNU-Prolog syntax
// 14 October 2002: Version 2.1.3
// - small bug fix in the incremental drawing of non finite domain variable


// To do:
// - adopt the OaDymPpac generic trace format
// - adopt the OaDymPpac event handling Java classes to support external viewers


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;

public class CLPGUI extends Frame implements WindowListener, ActionListener, ItemListener{

    static final int timeMax=10000;
    static final int buttonsMax=50;
    static final int charSize=8;
    static final Color domainsColor=new Color(100,100,250);
    static final Color domainsLightColor=new Color(150,150,250);
    //    static final Color translucentGray=new Color(100,100,100,100); // too expensive

    Panel p;
    Domains3D domains3D;
    Domains2D domains2D;
    SearchTree3D searchTree3D;
    SearchTree root, current, highlight;
    int timestamp; 
    String info;
    boolean recompute, recomputeAll;
    SearchTree jumpto;
    SearchTreeList recomputation;
    
    TextField machine = new TextField("localhost",9);
    TextField port = new TextField("27311",5);
    TextField goal = new TextField("queens(6,L)",40);

    Button connect = new Button("CLP connect");
    Button disconnect = new Button("Disconnect");
    Button clear = new Button("Clear");
    Button backtrack = new Button("Backtrack");
    Button backtrackInteraction = new Button("Backtrack to last interaction");
    Button continueRecomputation = new Button("Continue");
    Button continueRecomputationBacktracking = new Button("Allow backtracking above recomputation node");
    Button execute = new Button("Execute:");
    Button[] goals;

    Menu options = new Menu("Viewers");
    MenuItem showDomains = new MenuItem("3D finite domains");
    MenuItem show2Domains = new MenuItem("2D finite domains");
    MenuItem showSearchTree = new MenuItem("Search tree");
    MenuItem printSearchTree = new MenuItem("Print search tree");
    CheckboxMenuItem backtrackTrace = new CheckboxMenuItem("Trace backtracking", false);
    CheckboxMenuItem domainValues = new CheckboxMenuItem("Show domain values", false);
    CheckboxMenuItem domainInterval = new CheckboxMenuItem("Show domain intervals", false);

    Socket socket;
    PrintWriter out=null;
    BufferedReader in=null;

    String[] names;
    SearchTree[] executedGoals;
    int[][] domains;
    int nbvar;
    int time;
    int ngoals;
    int maxSize;
    int vMin, vMax;

    boolean refresh=false;

    public CLPGUI()
    {
	System.out.println("CLPGUI v2.1.3 by INRIA, Franois Fages (C) 2001-2002. Java process.");
	setSize(700,250);
	setBackground(Color.lightGray);

	setLayout(new BorderLayout());

	options.add(showSearchTree);
	options.add(printSearchTree);
	options.add(showDomains);
	options.add(show2Domains);
	options.add(domainValues);
	options.add(domainInterval);
	MenuBar menuBar=new MenuBar();
	menuBar.add(options);
	setMenuBar(menuBar);

	Panel buttons=new Panel();
	//buttons.setBackground(Color.white);
	buttons.setLayout(new BorderLayout());

	p=new Panel(); // Panel of connections
	p.setLayout(new FlowLayout(FlowLayout.LEFT));
	p.add(connect); 
	p.add(new Label("on"));
	p.add(machine);
	p.add(new Label("port"));
	p.add(port);
	disconnect.setEnabled(false);
	p.add(disconnect);
	buttons.add(p,BorderLayout.NORTH);

	p=new Panel(); // Panel of commands
	p.setLayout(new FlowLayout(FlowLayout.LEFT));
	p.add(execute);
	p.add(goal);
	p.add(backtrack);
	p.add(backtrackInteraction);
	p.add(clear);
	buttons.add(p,BorderLayout.CENTER);
	p=new Panel(); // Panel of extra commands
	p.setLayout(new FlowLayout(FlowLayout.RIGHT));
	p.add(continueRecomputation);
	p.add(continueRecomputationBacktracking);
	continueRecomputationBacktracking.setEnabled(false);
	buttons.add(p,BorderLayout.SOUTH);

	add(buttons,BorderLayout.NORTH);

	p=new Panel(); // panel of interactive goals
	//p.setBackground(Color.lightGray);
	p.setLayout(new FlowLayout(FlowLayout.LEFT));
	add(p,BorderLayout.CENTER);

	connect.addActionListener(this);
	disconnect.addActionListener(this);
        clear.addActionListener(this);
	backtrack.addActionListener(this);
	backtrackInteraction.addActionListener(this);
	continueRecomputation.addActionListener(this);
	continueRecomputationBacktracking.addActionListener(this);
	execute.addActionListener(this);
	printSearchTree.addActionListener(this);
	showDomains.addActionListener(this);
	show2Domains.addActionListener(this);
	showSearchTree.addActionListener(this);

	backtrackTrace.addItemListener(this);
	domainValues.addItemListener(this);
	domainInterval.addItemListener(this);

/*	domains2D=new Domains2D();
	domains2D.setTitle("CLPGUI Finite Domains 2D Viewer");
	domains2D.addWindowListener(domains2D);
	domains2D.show();

	domains3D=new Domains3D();
	Frame3D window=new Frame3D(domains3D);
	window.setTitle("CLPGUI Finite Domains 3D Viewer");
      	window.addWindowListener(window);
	window.show();

	searchTree3D=new SearchTree3D();
	window=new Frame3D(searchTree3D);
	window.setTitle("CLPGUI Search Tree Viewer");
      	window.addWindowListener(window);
	window.show(); */

	init();

    }

    void init() {
	initVariables();
	root=new SearchTree("root");
	current=root;
	highlight=null;
	timestamp=1;
	recompute=false;
	continueRecomputationBacktracking.setEnabled(false);
	jumpto=null;
    }

    void initVariables() {
	info="";
	executedGoals=new SearchTree[timeMax];
	goals = new Button[buttonsMax];
	maxSize=0;
	vMin=10000; vMax=-10000;
	ngoals=0;
	time=0;
	nbvar=0;
	p.removeAll();
	repaint();
    }

    void clear() {
	init();
	if (domains2D!=null) domains2D.clear();
	if (domains3D!=null) domains3D.clear();
	if (searchTree3D!=null) searchTree3D.clear();
    }

    void disconnect() {
        send("d");
        clear();
        socketClose();
        disconnect.setEnabled(false);
        connect.setEnabled(true);
        repaint();   
    }

    void closing(Canvas3D c) {
	if (c==domains3D) domains3D=null;
	if (c==searchTree3D) searchTree3D=null;
    }

    // recomputation 
    synchronized void recompute() {
	if (jumpto!=null & jumpto!=root) {
	    if (ancestorRecompute(jumpto)) { // bug in the current implementation...
		System.err.println("Error: cannot recompute below a recomputation node.");
		jumpto=null;
		return;
	    }
	    recompute=true;
	    continueRecomputationBacktracking.setEnabled(true);
	    initVariables();
	    recomputation=null;
	    while (jumpto.label.startsWith("recompute ")) {
		jumpto=jumpto.father;
	    }
	    current=jumpto;
	    timestamp=1;
	    while (current!=root) {
		recomputation=new SearchTreeList(current,recomputation);
		current=current.father;
	    }
	    send("j");
	    sendJumpto(jumpto.father);
	    send("l");
	    if (jumpto.father.property.endsWith("success")) {
		send(jumpto.label);
	    }
	    else {
		send(""+jumpto.father.childNumber(jumpto));
	    }
	    jumpto=null;
	}
	else
	    System.err.println("Error recompute null.");
    }
    
    synchronized void sendJumpto(SearchTree t) {
	if (t==root) return;
	if (!t.label.startsWith("recompute")) 
	    sendJumpto(t.father);
	else {
	    sendJumpto(t.father.father);
	    send("m");
	}
	if (t.father.property.endsWith("success")) {
	    if (t.label.endsWith("continue"))
		send("c");
	    else
		send(t.label);
	}
	else 
	    send(""+t.father.childNumber(t));
    }
    
    boolean ancestor(SearchTree t) {
	SearchTree p=current;
	while (p!=root) {
	    if (p==t) return true;
	    else p=p.father;
	}
	return (t==root);
    }

    boolean ancestorRecompute(SearchTree p) {
  	while (p!=root) {
 	    if (p.label.startsWith("recompute ")) return true;
 	    else p=p.father;
  	}
 	return false;
    }

    public synchronized void paint(Graphics g) {
    }

    public void refresh() {
     	p.doLayout();
       	doLayout();
  	// pack();
	refresh=true;
    }


    public void itemStateChanged(ItemEvent e) {
	String s=e.getItem().toString();
	if (s.equals("Show domain values")) {
	    if (domainValues.getState()) {
		domainInterval.setState(false);
		send("sv");
	    }
	    else send("sz");
	}
	else if (s.equals("Show domain intervals")) {
	    if (domainInterval.getState()) {
		domainValues.setState(false);
		send("sl");
	    }
	    else send("sz");
	}
    }
    
    public void actionPerformed(ActionEvent e)  
    {
	if (e.getSource()==connect) {
	    socketInit();
	    connect.setEnabled(false);
	    disconnect.setEnabled(true);
	    repaint();
	}
	else if (e.getSource()==disconnect) {
	    disconnect();
	}
	else if (e.getSource()==clear) {
	    send("e");
	    clear();
	}
	else if (e.getSource()==domainValues) {
	}
	else if (e.getSource()==backtrack) {
	    send("b");
	}
	else if (e.getSource()==backtrackInteraction) {
	    send("a");
	}
	else if (e.getSource()==continueRecomputation) {
	    send("c");
	}
	else if (e.getSource()==continueRecomputationBacktracking) {
	    continueRecomputationBacktracking.setEnabled(false);
	    send("k");
	}
	else if (e.getSource()==execute) {
	    send(goal.getText());
	}
	else if (e.getSource()==showSearchTree) {
	    searchTree3D=new SearchTree3D();
	    Frame3D window=new Frame3D(searchTree3D);
	    window.setTitle("CLPGUI Search Tree Viewer");
	    window.addWindowListener(window);
	    window.show();
	}
	else if (e.getSource()==showDomains) {
	    domains3D=new Domains3D();
	    Frame3D window=new Frame3D(domains3D);
	    window.setTitle("CLPGUI Finite Domains 3D Viewer");
	    window.addWindowListener(window);
	    window.show();
	}
	else if (e.getSource()==show2Domains) {
	    domains2D=new Domains2D();
	    domains2D.setTitle("CLPGUI Finite Domains 2D Viewer");
	    domains2D.addWindowListener(domains2D);
	    domains2D.show();
	}
	else if (e.getSource()==printSearchTree) {
	    root.prettyPrint();
	    System.out.println();
	}
	else {
	    ((Button)(e.getSource())).setEnabled(false);
	    send(e.getActionCommand());	    
	}

	//	paint(getGraphics());  
    }
    
    public void windowClosing(WindowEvent e)
    {
	exit();
    }
    public void windowClosed(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e) {}
    public void windowActivated(WindowEvent e) {}
    public void windowDeactivated(WindowEvent e) {}
    public void windowOpened(WindowEvent e) {}

    // Communication with the CLP server through sockets

    void socketInit()
    {
	try {
	    socket=new Socket(machine.getText(),Integer.parseInt(port.getText()));
	    out= new PrintWriter(socket.getOutputStream(), true);
	    in= new BufferedReader(new InputStreamReader(socket.getInputStream()));
	}
	catch (Exception e) {
	    System.err.println("Socket opening error: "+e.getMessage());
	}
	Interpreter interpreter= new Interpreter(this);
	interpreter.start();
    }

    void socketClose () {
	try {
	    in.close();
	    out.close();
	    socket.close();
	}
	catch (Exception x) {}
	System.exit(0);
    }

    void exit() {
	socketClose();
    }

    void send(String s)
    {
	while (s.endsWith(".")) s=s.substring(0,s.length()-2);
	s=s+".";
	try {
	    //System.err.println("Sending "+s);
	    out.println(s);
	} catch (Exception x) {System.err.println("Error sending "+s);}
    }

    String receive()
    {
	String s="";
	try {
	    s=in.readLine();
	    //System.err.println("Received "+s);
	} catch (Exception x) {System.err.println("Error reading socket.");}
	return s;
    }

    int receiveInt() {
	return receiveInt("Error reading integer ");
    }

    int receiveInt(String m) {
	String z=receive();
	try {
	    return Integer.parseInt(z);
	}
	catch (NumberFormatException x) {
	    System.err.println(m+z);
	    return 0;
	}
    }

    int intOf(String z, int i, int j) {
	z=z.substring(i,j);
	try {
	    return Integer.parseInt(z);
	}
	catch (NumberFormatException x) {
	    System.err.println("Error reading domain integer "+z);
	    return 0;
	}
    }

    int indexDomain(String z, int i) {
	int j=z.indexOf('.',i);
	int k=z.indexOf(':',i);
	if (j<0) {
	    if (k<0) return z.indexOf(')',i);
	    else return k;
	}
	else {
	    if (k<0) return j;
	    else return Math.min(j,k);
	}
    }

    public static void main (String args[])
    {
	CLPGUI window= new CLPGUI();
	window.setTitle("Constraint Logic Programming GUI               2001,2002 INRIA projet Contraintes");
	window.addWindowListener(window);
	try {
	    window.port.setText(args[0]);
	} catch (Exception x) {};

	window.show();
    }

    //////////////////////////////////////////////////////////////////
    // Inner class for interpreting the commands read on the socket //
    //////////////////////////////////////////////////////////////////

    class Interpreter extends Thread {
	CLPGUI g;
    
	Interpreter(CLPGUI a) {
	    g=a;
	}

	public void run() {
	    while (socket!=null) {
		try {
		    interpret();
		} catch (Exception x) {System.err.println("Error in run: "+x.toString());}
	    }
	}

	public synchronized void interpret() {
	    String s;
	    s=receive();
	    if (s==null) return;

	    if (s.equals("t")) { // timestamp
		int t;
		t=receiveInt();

		if (recompute) {
		    if (timestamp>t) {
			System.err.println("Error: backtracking during recomputation from depth "+timestamp+" to "+t);
			recompute=false;
		    }
		    else
			return;
		}

		if (timestamp>t) { // backtracking... show_domains has been executed
		    if (domains2D!=null) {
			domains2D.paint(domains2D.getGraphics());
		    }
		    refresh=true;
		}
		while (timestamp>t) {
		    s=current.label;
		    if (current.father.property.endsWith("success")) {
			for (int i=0; i<ngoals; i++) {
			    if (s.equals(goals[i].getLabel())) {
				goals[i].setEnabled(true);
				repaint();
				break;
			    }
			}
		    }
		    if (!current.label.endsWith("continue")) timestamp--; // continue nodes have no CLP counterpart
		    current=current.father;
		}
		if (!backtrackTrace.getState() & !current.label.startsWith("recompute")){
		    while (time>0 && executedGoals[time-1]!=current) {
			refresh=true;
			executedGoals[time]=null;
			time--;
		    }
		}
	    }
	    else if (s.equals("n") || s.equals("o") || s.equals("c") || s.equals("continue")) { // new node in the search tree
		if (recompute) {
		    if (!s.equals("continue")) s=receive();
		    if (recomputation==null) {
			System.err.println("Error: recomputation beyond stop point, executing "+s);
			recompute=false;
		    }
		    else if (s.equals(recomputation.head.label)) {
			current=recomputation.head;
			recomputation=recomputation.tail;
			if (!s.equals("continue")) timestamp++;
		    }
		    else {
			System.err.println("Error: recomputation in wrong branch "+s);
			recompute=false;
		    }
		}
		else {
		    SearchTree r=new SearchTree();
		    if (s.equals("o")) r.property="show";
		    if (s.equals("continue")) 
			r.label="continue";
		    else {
			s=receive();
			r.label=s;
			timestamp++;
		    }
		    current.addChild(r);
		    current=r;
		}
	    }
	    else if (s.equals("r")) { // end recomputation
		//System.err.println("end recomputation "+current.label);
		if (!recompute) {
		    System.err.println("Error in recomputation");
		}
		else if (recomputation!=null) {
		    System.err.println("Error: recomputation stopped before "+recomputation.head.label);
		    recomputation=null;
		}
		
		recompute=false;
		SearchTree r=new SearchTree(s);
		r.label="recompute "+current.label;
		current.addChild(r);
		current=r;
		timestamp++;		
	    }
	    else if (s.equals("z") || s.equals("v") || s.equals("l")) { // finite domains
		executedGoals[time]=current;
		
		if (domains3D!=null) domains3D.eraseNames();
		
		int d;
		d=receiveInt("Error reading number of variables ");

		if (nbvar!=d) {
		    System.err.println("Warning: variables out of date");
		    refresh=true;
		    nbvar=d;
		}

		if (domains2D!=null) domains2D.clear();

		for (int v=0; v<nbvar; v++) {
		    d=receiveInt("Error reading domain size ");
		    domains[v][time]=d;
		    if (d>maxSize) maxSize=d;
		    if (domains3D!=null) {
			domains3D.parameter(v);
			domains3D.drawDomainVar();
		    }
		    if (d>0) {
			if (s.equals("v")) {
			    int n1;
			    if (d==1) {
				n1=receiveInt("Error reading single domain value ");
				if (domains2D!=null) {
				    if (n1<vMin) vMin=n1;
				    else if (n1>vMax) vMax=n1;
				    domains2D.inDomain(v,n1);
				}
				if (domains3D!=null) {
				    domains3D.k=0;
				    domains3D.drawValue(""+n1);
				}
			    }
			    else { // domain as list of intervals #_45(1:3:5..10)
				String z;
				z=receive();
				//System.err.println("Domain "+z);
				int i,j,n2;
				i=z.indexOf('(')+1;
				if (i<=0) 
				    System.err.println("Error reading domain "+z);
				else {
				    int k=0;
				    do {
					j=indexDomain(z,i);
					n1=intOf(z,i,j);
					if (z.charAt(j)=='.') { // interval n1..n2
					    i=j+2;
					    j=indexDomain(z,i); // : or )
					    n2=intOf(z,i,j);
					    if (domains2D!=null) {
						if (n1<vMin) vMin=n1;
						if (n2>vMax) vMax=n2;
						domains2D.inDomain(v,n1,n2);
					    }
					    if (domains3D!=null) {
						for (int val=n1;val<=n2;val++) {
						    domains3D.k=k++;
						    domains3D.drawValue(""+val);
						}
					    }
					    i=j+1;
					}
					else if (z.charAt(j)==':') {
					    if (domains2D!=null) {
						if (n1<vMin) vMin=n1;
						if (n1>vMax) vMax=n1;
						domains2D.inDomain(v,n1);
					    }
					    if (domains3D!=null) {
						domains3D.k=k++;
						domains3D.drawValue(""+n1);
					    }
					    i=j+1;
					}
				    }
				    while (z.charAt(j)!=')');
				}
			    }
			}
			else if (s.equals("l") & d>0) {
			    int min=receiveInt("Error reading interval minimum ");
			    int max=receiveInt("Error reading interval maximum ");
			    
			    if (domains2D!=null) {
				if (min<vMin) vMin=min;
				if (max>vMax) vMax=max;
				domains2D.inDomain(v,min,max);
			    }
			    if (domains3D!=null) {
				domains3D.k=0;
				domains3D.drawValue(""+min);
				domains3D.k=d-1;
				domains3D.drawValue(""+max);
			    }
			}
		    }
		}
		if (nbvar>0) time++;
		if (domains3D!=null) {
		    //domains3D.drawGoalName(current,domains3D.kv*nbvar,kt*time+kt,0);
		    domains3D.paint(domains3D.getGraphics()); //incremental drawing with refresh=false
		}
		//if (domains2D!=null) domains2D.paint(domains2D.getGraphics());
	    }
	    else if (s.equals("w")) { // show node
		if (recompute) return;
		current.property="show"+current.property;
	    }
	    else if (s.equals("s")) { // success
		if (recompute) return;
		current.property=current.property+"success";
		//refresh=true;
		if (searchTree3D!=null) searchTree3D.paint(searchTree3D.getGraphics());		     
		if (domains3D!=null) domains3D.paint(domains3D.getGraphics());
		if (domains2D!=null) domains2D.paint(domains2D.getGraphics());
	    }
	    else if (s.equals("b")) { // new button
		s=receive();
		Button button=new Button(s);
		goals[ngoals++]=button;
		p.add(button);
		button.addActionListener(g);
		refresh();
	    }
	    else if (s.equals("a")) { // variables names
		nbvar=receiveInt("Error reading number of variables ");
		names=new String[nbvar];
		domains=new int[nbvar][timeMax];
		time=0;
		for (int i=0; i<nbvar; i++) {
		    s=receive();
		    names[i]=s;
		}
		refresh=true;
	    }
	    else if (s.equals("nt")) { // no trace backtracked goals
		backtrackTrace.setState(false);
	    }
	    else if (s.equals("tr")) { // trace backtracked goals
		backtrackTrace.setState(true);
	    }
	    else if (s.equals("sn")) { // show no domain
		domains3D=null;
	    }
	    else if (s.equals("sv")) { // show domain values
		domainValues.setState(true);
		domainInterval.setState(false);
	    }
	    else if (s.equals("sl")) { // show domain intervals
		domainInterval.setState(true);
		domainValues.setState(false);
	    }
	    else if (s.equals("sz")) { // show domain sizes
		domainInterval.setState(false);
		domainValues.setState(false);
	    }
	    else if (s.equals("e")) // clear
		clear();
	    else if (s.equals("i")) { // info
		s=receive()+"="+receive();
		if (s.startsWith(info)) // otherwise that info arrives too late...
		    info=s; 
	    }
	    else System.err.print("Ignoring reception: "+s+" ");
	}
    }

    ////////////////////////////////////////
    // Simple 2D viewer of finite domains //
    ////////////////////////////////////////

    class Domains2D extends Frame implements WindowListener{
	int width=500;
	int height=500;
	int di,dj;
	Graphics gMem;    // double buffering 
	Image memory;

	Domains2D() {
	    setSize(width,height);
	    init();
	}

	void init() {
	    di=(height-60)/(nbvar+1);
	    if (di<1) di=1;
	    dj=(width-60)/(vMax-vMin+1);
	    if (dj<1) dj=1;
	}

	void resize() {
	    memory = createImage(width,height);
	    gMem = memory.getGraphics();
	    gMem.setFont(new Font("monospaced",Font.PLAIN,12));
	    clear();
	}

	void clear() {
	    init();
	    gMem.setColor(Color.white);
	    gMem.fillRect(0, 0, width, height);
	    gMem.setColor(Color.lightGray);
	    for (int v=0;v<nbvar;v++) {
		gMem.drawString(names[v],5,60+v*di+di);
		gMem.drawLine(50,60+v*di,50+(vMax-vMin+1)*dj,60+v*di);
	    }
	    gMem.drawLine(50,60+nbvar*di,50+(vMax-vMin+1)*dj,60+nbvar*di);
	    for (int v=vMin;v<=vMax;v++) {
		gMem.drawString(""+v,50+(v-vMin)*dj,45);
		gMem.drawLine(50+(v-vMin)*dj,60,50+(v-vMin)*dj,60+nbvar*di);
	    }
	    gMem.drawLine(50+(vMax+1-vMin)*dj,60,50+(vMax+1-vMin)*dj,60+nbvar*di);
	    gMem.setColor(Color.red);
	}

	void inDomain(int v, int d) {
	    gMem.fillRect(50+(d-vMin)*dj,60+v*di,dj,di);
	}

	void inDomain(int v, int min, int max) {
	    gMem.fillRect(50+(min-vMin)*dj,60+v*di,(max-min+1)*dj,di);
	}

	public synchronized void paint(Graphics g) {
	    if (width!=getSize().width || height!=getSize().height) {
		width=getSize().width;
		height=getSize().height;
		memory=null;
	    }
	    if (memory==null) resize();
	    g.drawImage(memory,0,0,this);
	}

	public void windowClosing(WindowEvent e)
	{
	if (domains2D==this) domains2D=null;
	    dispose();
	}
	public void windowClosed(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowActivated(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}
	public void windowOpened(WindowEvent e) {}
    }


    ///////////////////////////////////////////////
    // A class for frames containing a 3D canvas //
    ///////////////////////////////////////////////

    class Frame3D extends Frame implements WindowListener{
	int width=500;
	int height=500;
	Canvas3D figure;
	
	Frame3D(Canvas3D c) {
	    setSize(width,height);
	    figure=c;
	    setMenuBar(c.menuBar);

	    figure.resized(width,height-30);
	    add(figure);
	}

	public synchronized void paint(Graphics g) {
	    if (width!=getSize().width || height!=getSize().height) {
		width=getSize().width;
		height=getSize().height;
		figure.resized(width,height-30);
	    }
	}

	public void windowClosing(WindowEvent e)
	{
	    closing(figure);
	    dispose();
	}
	public void windowClosed(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowActivated(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}
	public void windowOpened(WindowEvent e) {}
    }

    ///////////////////////////
    // A class for 3D canvas //
    ///////////////////////////

    class Canvas3D extends Canvas implements MouseListener, MouseMotionListener, ActionListener, ItemListener{
	MenuBar menuBar;
	Menu display;
	double zoomX, zoomY;
	CheckboxMenuItem rotate, translate, zoom;
	MenuItem center;

	int width=500;
	int height=450;
	Graphics gMem;    // double buffering 
	Image memory;

	double oX;
	double oY;
	int lastX,lastY,firstX,firstY;
	
	int[] x=new int[6]; // temporary points
	int[] y=new int[6];

	double aY,aZ;
	double sinRZ, cosRZ, sinRY, cosRY;
	
	static final double pi6=Math.PI/6;
	static final double pi4=Math.PI/4;
	static final double pi2=Math.PI/2;

	Canvas3D() {
	    setSize(width,height);
	    memory=null;
	    addMouseListener(this);
	    addMouseMotionListener(this);

	    menuBar=new MenuBar();
	    display=new Menu("Geometry");
	    menuBar.add(display);

	    rotate=new CheckboxMenuItem("Rotate",true);
	    translate=new CheckboxMenuItem("Translate",false);
	    zoom=new CheckboxMenuItem("Zoom",false);
	    center=new MenuItem("Center");
	    center.addActionListener(this);
	    rotate.addItemListener(this);
	    translate.addItemListener(this);
	    zoom.addItemListener(this);
	    display.add(center);
	    display.add(rotate);
	    display.add(translate);
	    display.add(zoom);

	    zoomX=1; // can't use center() because overiding...
	    zoomY=1;
	    oX=0;
	    oY=0;
	    aY=0;
	    aZ=0;
	    initRotate();
	}

	void center() {
	    zoomX=1;
	    zoomY=1;
	    oX=0;
	    oY=0;
	    aY=0;
	    aZ=0;
	    initRotate();
	    refresh=true;
	    repaint();
	}

	void resized(int w, int h) {
	    width=w;
	    height=h;
	    setSize(width,height);
	    memory=null;
	    refresh=true;
	}

	// Rotation

	void initRotate() {
	    sinRZ=Math.sin(aZ);
	    cosRZ=Math.cos(aZ);
	    sinRY=Math.sin(aY);
	    cosRY=Math.cos(aY);
	}

	// Projection from xyz into ij
	int i(double x, double y, double z) {
	    return (int)(oX+zoomX*(y*cosRZ+x*sinRZ));
	}
	
	int j(double x, double y, double z) {
	    return (int)(oY+zoomY*(-z*cosRY+(x*cosRZ-y*sinRZ)*sinRY));
	}

	// Inverse projection from ij into xy when z=0
	double x(int ii, int jj) {
	    double i=(ii-oX)/zoomX;
	    double j=(jj-oY)/zoomY;
	    return (i-y(ii,jj)*cosRZ)/sinRZ;
	}
	
	double y(int ii, int jj) {
	    double i=(ii-oX)/zoomX;
	    double j=(jj-oY)/zoomY;
	    return i*cosRZ-j*sinRZ/sinRY;
	}

	// Inverse projection from ij into xy knowing z
	double x(int ii, int jj, double z) {
	    double i=(ii-oX)/zoomX;
	    double j=(jj-oY)/zoomY;
	    return (i-y(ii,jj,z)*cosRZ)/sinRZ;
	}
	
	double y(int ii, int jj, double z) {
	    double i=(ii-oX)/zoomX;
	    double j=(jj-oY)/zoomY;
	    return i*cosRZ-(j+z*cosRY)*sinRZ/sinRY;
	}

	// Menus
	public void actionPerformed(ActionEvent e) {
	    if (e.getSource()==center) {
		center();
	    }
	}

	public void itemStateChanged(ItemEvent e) {
	    String s=e.getItem().toString();
	    
	    if (s.equals("Rotate") && rotate.getState()) {
		translate.setState(false);
		zoom.setState(false);
	    }
	    else if (s.equals("Translate") && translate.getState()) {
		rotate.setState(false);
		zoom.setState(false);
	    }
	    else if (s.equals("Zoom") && zoom.getState()){
		translate.setState(false);
		rotate.setState(false);		
		zoom.setState(true);		
	    }
	    //else return;

	    refresh=true;
	    repaint();
	}
    
	// Mouse events
	public synchronized void mouseClicked(MouseEvent e)
	{
	    if (rotate.getState()) {
		rotate.setState(false);
		translate.setState(true);
		zoom.setState(false);
	    }
	    else if (translate.getState()) {
		rotate.setState(false);
		translate.setState(false);
		zoom.setState(true);
	    }
	    else if (zoom.getState()) {
		rotate.setState(true);
		translate.setState(false);
		zoom.setState(false);
	    }
	}

	public synchronized void mousePressed(MouseEvent e)
	{
	    lastX=e.getX();
	    lastY=e.getY();
	    firstX=lastX;
	    firstY=lastY;
	}

	public void mouseReleased(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}

	public synchronized void mouseDragged(MouseEvent e)
	{
	    if (rotate.getState()) {
		double raZ=aZ;
		double raY=aY;
		aZ=aZ+(e.getX()-lastX)/300.0;
		if (aZ<-pi2 || aZ>0) aZ=raZ;
		aY=aY+(e.getY()-lastY)/300.0;
		if (aY<0 || aY>pi2) aY=raY;
		sinRZ=Math.sin(aZ);
		cosRZ=Math.cos(aZ);
		sinRY=Math.sin(aY);
		cosRY=Math.cos(aY);
		lastX=e.getX();
		lastY=e.getY();
	    }
	    else if (translate.getState()) {
		oX=oX+e.getX()-lastX;
		oY=oY+e.getY()-lastY;
		lastX=e.getX();
		lastY=e.getY();
	    }
	    else {
		double dZoomX=(e.getX()-lastX)*0.01;
		double dZoomY=(e.getY()-lastY)*0.01;
		// zoom with a translation that preserves the coordinate of the first pressed point
		// a negative zoom reverses the figure
  		if (zoomX+dZoomX>0.01 || zoomX+dZoomX<-0.01) {
		    oX=firstX-(firstX-oX)*(1+dZoomX/zoomX);
		    zoomX=zoomX+dZoomX;
		    lastX=e.getX();
		}
  		if (zoomY+dZoomY>0.02 || zoomY+dZoomY<-0.02) {
		    oY=firstY-(firstY-oY)*(1+dZoomY/zoomY);
		    zoomY=zoomY+dZoomY;
		    lastY=e.getY();
		}
	    }
	    refresh=true;
	    paint(getGraphics());
	}

	public void mouseMoved(MouseEvent e) {}

	public synchronized void paint(Graphics g)     
	{
	    if (memory==null) {
		memory = createImage(width,height);
		gMem = memory.getGraphics();
		gMem.setFont(new Font("monospaced",Font.PLAIN,12));
		refresh=true;
	    }

	    if (refresh) {
		gMem.setColor(Color.white);
		gMem.fillRect(0, 0, width, height);
	    }
	}
    }

    ////////////////////////////////
    // A 3D Finite Domains viewer //
    ////////////////////////////////

    class Domains3D extends Canvas3D {
	Menu options=new Menu("Options");

	int scaleX, scaleY, scaleZ;
	int kv, kt, ks, rv, rt, rs;
	int varHighlight=-1;
	
	int v,k,d;

	Domains3D(){
	    super();
	    options.add(backtrackTrace);
	    menuBar.add(options);

	    clear();
	}

	void clear() {
	    kv=40;
	    kt=40;
	    ks=20;
	    center();	    
	}

	void center() {
	    super.center();
	    oX=width/2-50;
	    oY=10+height/3;
	    aZ=-pi4;
	    aY=pi6;
	    initRotate();
	    repaint();
	}

	void resized(int width, int height) {
	    super.resized(width,height);
	    oX=width/2-30;
	    oY=10+height/3;
	    scaleX=width/2+100;
	    scaleY=width/2+100;
	    scaleZ=height/3;
	    autoscale();
	    refresh=true;
	}
	
	void autoscale() {
	    if (kv*nbvar<width/3) kv=40;
	    if (ks*maxSize<height/4) ks=20;
	    if (kt*time<width/3) kt=20;

	    if (kv*nbvar>width/2 || kv==1) kv=width/(2*(nbvar+1));
	    if (ks*maxSize>height/3 || ks==1) ks=height/(3*(maxSize+1));
	    if (kt*time>width/2 || kt==1) kt=width/(2*(time+1));
	    if (kv<1) kv=1;
	    if (ks<1) ks=1;
	    if (kt<1) kt=1;
	}

	public synchronized void mouseMoved(MouseEvent e)
	{
	    double x=x(e.getX(),e.getY())/kv;
	    double y=y(e.getX(),e.getY())/kt;
	    
	    if (y>0 & y<time+10 & x>0 & x<nbvar) {
		if (y>time) {
		    if (varHighlight!=(int)x) {
			refresh=true;
			varHighlight=(int)x;
			info="";
		    }
		}
		else {
		    varHighlight=-1;
		    info="";
		}
		if (x>nbvar-1 & y<time) {
		    highlight=executedGoals[(int)y]; 
		    refresh=true;
		    if (searchTree3D!=null) searchTree3D.paint(searchTree3D.getGraphics());
		}
		else if (highlight!=null) {
		    refresh=true;
		    //highlight=null;
		    if (searchTree3D!=null) searchTree3D.paint(searchTree3D.getGraphics());
		}
	    }
	    else {
		varHighlight=-1;
		info="";
	    }
	    paint(getGraphics());
	}
	
	public synchronized void drawAll()
	{
		gMem.setColor(Color.gray);
		gMem.drawString("0",i(-10,0,0),j(-10,0,0));
		gMem.drawLine(i(0,0,0),j(0,0,0),i(scaleX,0,0),j(scaleX,0,0));
		gMem.drawLine(i(0,0,0),j(0,0,0),i(0,scaleY,0),j(0,scaleY,0));
		gMem.drawLine(i(0,0,0),j(0,0,0),i(0,0,scaleZ),j(0,0,scaleZ));
		gMem.drawString(nbvar+" variables",i(scaleX,0,0),j(scaleX,0,0));
		gMem.drawString("time "+time,i(0,scaleY,0),j(0,scaleY,0));
		gMem.drawString("domain size "+maxSize,i(0,0,scaleZ),j(0,0,scaleZ));

		for (int t=0; t<time; t++) {
		    rt=kt*t;

		    Color col;
		    if (ancestor(executedGoals[t]))
			col=domainsColor;
		    else
			col=domainsLightColor;

		    for (int v=0; v<nbvar; v++) {
			rv=kv*v;
			rs=ks*domains[v][t];
			// top facet
			x[0]=i(rv+kv,rt,rs);
			y[0]=j(rv+kv,rt,rs);
			x[1]=i(rv,rt,rs);
			y[1]=j(rv,rt,rs);
			x[2]=i(rv,rt+kt,rs);
			y[2]=j(rv,rt+kt,rs);
			if (v<nbvar-1 && t<time-1 && rs<=ks*domains[v+1][t] && rs<=ks*domains[v][t+1]) {
			    // top facet only
			    x[3]=i(rv+kv,rt+kt,rs);
			    y[3]=j(rv+kv,rt+kt,rs);
			    gMem.setColor(col);
			    gMem.fillPolygon(x,y,4);
			    gMem.setColor(Color.black);
			    gMem.drawPolygon(x,y,4);
			}
			else {
			    // left and right facets added
			    x[3]=i(rv,rt+kt,0);
			    y[3]=j(rv,rt+kt,0);
			    x[4]=i(rv+kv,rt+kt,0);
			    y[4]=j(rv+kv,rt+kt,0);
			    x[5]=i(rv+kv,rt,0);
			    y[5]=j(rv+kv,rt,0);
			    gMem.setColor(col);
			    gMem.fillPolygon(x,y,6);
			    gMem.setColor(Color.black);
			    gMem.drawPolygon(x,y,6);
			    x[1]=i(rv+kv,rt+kt,rs); // missing edges
			    y[1]=j(rv+kv,rt+kt,rs);
			    gMem.drawLine(x[1],y[1],x[4],y[4]);
			    gMem.drawLine(x[1],y[1],x[0],y[0]);
			    gMem.drawLine(x[1],y[1],x[2],y[2]);
			}
		    }
		    drawGoalName(executedGoals[t],kv*nbvar,kt*t+kt,0);
		}
	}

	public synchronized void drawGoalName(SearchTree t, double x, double y, double z) {
	    
	    if (t.property.endsWith("success")) {
		gMem.setColor(Color.red);
		gMem.drawString("+",i(x,y,z)-charSize, j(x,y,z));
	    }
	    
	    if (t==current) {
		gMem.setColor(Color.blue);
	    }
	    else if (t==root) {
		gMem.setColor(Color.green);
	    }
	    else if (t==highlight || t.property.startsWith("show")) {
		gMem.setColor(Color.red);
	    }
	    else return;
	    
	    int l; if (zoomX>0) l=t.label.length()*charSize; else l=charSize;
	    if (t.property.endsWith("success")) 
		gMem.drawString(t.label, i(x,y,z)-l-charSize, j(x,y,z));
	    else
		gMem.drawString(t.label, i(x,y,z)-l, j(x,y,z));
	}
    
	public synchronized void eraseNames()
	{
	    gMem.setColor(Color.white); // erases variable names and previous time
	    for (int v=0; v<nbvar; v++) {
		gMem.drawString(names[v],i(kv*(v+1),kt*time+20,0),j(kv*(v+1),kt*time+20,0));
	    }
	    gMem.drawString("time "+(time-1),i(0,scaleY,0),j(0,scaleY,0));
	    gMem.drawString("time "+(time+1),i(0,scaleY,0),j(0,scaleY,0));

	    gMem.setColor(Color.gray);
	    gMem.drawString("time "+time,i(0,scaleY,0),j(0,scaleY,0));

	}

	public void parameter(int var)
	{
	    v=var;
	    d=domains[v][time];
	    rt=kt*time;
	    rv=kv*v;
	    rs=ks*d;
	}

	public synchronized void drawDomainVar() 
	{
	    // top facet
	    x[0]=i(rv+kv,rt,rs);
	    y[0]=j(rv+kv,rt,rs);
	    x[1]=i(rv,rt,rs);
	    y[1]=j(rv,rt,rs);
	    x[2]=i(rv,rt+kt,rs);
	    y[2]=j(rv,rt+kt,rs);
	    x[3]=i(rv,rt+kt,0);
	    y[3]=j(rv,rt+kt,0);
	    x[4]=i(rv+kv,rt+kt,0);
	    y[4]=j(rv+kv,rt+kt,0);
	    x[5]=i(rv+kv,rt,0);
	    y[5]=j(rv+kv,rt,0);
	    gMem.setColor(domainsColor);
	    gMem.fillPolygon(x,y,6);
	    gMem.setColor(Color.black);
	    gMem.drawPolygon(x,y,6);
	    x[1]=i(rv+kv,rt+kt,rs); // missing edges
	    y[1]=j(rv+kv,rt+kt,rs);
	    gMem.drawLine(x[1],y[1],x[4],y[4]);
	    gMem.drawLine(x[1],y[1],x[0],y[0]);
	    gMem.drawLine(x[1],y[1],x[2],y[2]);
	}

	public synchronized void drawValue(String z)
	{
	    gMem.setColor(Color.white);
	    gMem.drawString(z, i(rv+kv/2, rt+kt, rs*k/d+rs/(2*d)), j(rv+kv/2, rt+kt, rs*k/d+rs/(2*d)));
	}

	public synchronized void drawNames()
	{
	    gMem.setColor(Color.black);
	    for (int v=0; v<nbvar; v++) {
		gMem.drawString(names[v],i(kv*(v+1),kt*time+20,0),j(kv*(v+1),kt*time+20,0));
	    }
	    if (varHighlight>=0 & varHighlight<nbvar) {
		gMem.setColor(Color.red);
		if (!info.startsWith(names[varHighlight]+"=")) {
		    info=names[varHighlight]+"=";
		    send("i("+names[varHighlight]+")");
		}
		gMem.drawString(info,i(kv*(varHighlight+1),kt*time+20,0),j(kv*(varHighlight+1),kt*time+20,0));
	    }
	}


	public synchronized void paint(Graphics g)     
	{
	    //if (kt>1 && kt*time>width) refresh=true;

	    super.paint(g);

	    if (refresh) {
		autoscale();
		drawAll();
		refresh=false;
	    }

	    drawNames();

	    g.drawImage(memory,0,0,this);
	}	    

	
    }

    /////////////////////////
    // Search Tree viewers //
    /////////////////////////

    class SearchTree3D extends Canvas3D implements ItemListener, ActionListener 
    {
	static final int NEAR=25;
	SearchTree fRoot, fCurrent,sRoot;

	Menu trees;
	CheckboxMenuItem freezeTree,showGoals, fsp, fw, tm3d;
	Menu nodes;
	MenuItem restoreHidden, removeRecomputed, recomputeFast, recomputeSideEffects, hideSubtree, showSubtree, showCurrentGoal;
	boolean noHide, noRecompute;

	boolean drawCenteredStrings=true;
	SearchTree goalHighLight=null;

	int im,jm;
	boolean clicked=false;
	boolean showCompleteGoal=false;
	String nearString;

	SearchTree3D()
	{
	    super();
	    rotate.setState(false);
	    zoom.setState(false);
	    translate.setState(true);

	    trees=new Menu("Trees");
	    nodes=new Menu("Clicked node");
	    menuBar.add(trees);
	    menuBar.add(nodes);

	    fw=new CheckboxMenuItem("Tree in a fixed width",true);
	    fsp=new CheckboxMenuItem("Tree with a fixed spacing",false);
	    tm3d=new CheckboxMenuItem("3D tree",false);
	    freezeTree=new CheckboxMenuItem("Freeze tree",false);
	    showGoals=new CheckboxMenuItem("Show all goals",false);
	    restoreHidden=new MenuItem("Restore hidden subtrees");
	    removeRecomputed=new MenuItem("Remove recomputed nodes");
	    fw.addItemListener(this);
	    fsp.addItemListener(this);
	    tm3d.addItemListener(this);
	    freezeTree.addItemListener(this);
	    showGoals.addItemListener(this);
	    restoreHidden.addActionListener(this);
	    removeRecomputed.addActionListener(this);
	    trees.add(fw);
	    trees.add(fsp);
	    trees.add(tm3d);
	    trees.add(showGoals);
	    trees.add(removeRecomputed);
	    trees.add(restoreHidden);
	    trees.add(freezeTree);

	    showCurrentGoal=new MenuItem("Show complete goal");
	    hideSubtree=new MenuItem("Hide subtree");
	    showSubtree=new MenuItem("Show subtree only");
	    recomputeFast=new MenuItem("Recompute");
      	    recomputeSideEffects=new MenuItem("Recompute with side effects");
	    recomputeFast.addActionListener(this);
	    recomputeSideEffects.addActionListener(this);
	    hideSubtree.addActionListener(this);
	    showSubtree.addActionListener(this);
	    showCurrentGoal.addActionListener(this);
	    nodes.add(recomputeFast);
	    //	    nodes.add(recomputeSideEffects);
	    nodes.add(hideSubtree);
	    nodes.add(showSubtree);
	    nodes.add(showCurrentGoal);
	    nodes.setEnabled(false);

	    clear();
	}

	void clear() {
	    center();
	    sRoot=root;
	    zoomY=0.7;
	}

	void center() {
	    super.center();
	    setO();
	    if (tm3d.getState()) {
		aZ=-pi4;
		aY=pi6;
	    }
	    else {
		aZ=0;
		aY=0;
	    }
	    initRotate();
	    repaint();
	}

	void setO() {
	    if (fsp.getState()) {
		oX=0;
		oY=0;
	    }
	    else if (fw.getState()){
		oX=width/2;
		oY=0;
	    }
	    else {
		oX=width/2;
		oY=0; //height/3;
	    }
	}

	void resized(int width, int height) {
	    super.resized(width,height);
	    setO();
	}

	public void itemStateChanged(ItemEvent e) {
	    String s=e.getItem().toString();

	    if (s.equals("Tree with a fixed spacing")) {
		fsp.setState(true);
		fw.setState(false);
		tm3d.setState(false);
		setO();
	    }
	    else if (s.equals("Tree in a fixed width")){
		fsp.setState(false);
		tm3d.setState(false);
		fw.setState(true);
		setO();
	    }
	    else if (s.equals("3D tree")){
		fsp.setState(false);
		tm3d.setState(true);
		fw.setState(false);
	    }
	    center();
	    super.itemStateChanged(e);
	    repaint();
	}

	public synchronized void actionPerformed(ActionEvent e) {
	    clicked=false;
	    nodes.setEnabled(false);
	    showCompleteGoal=false;

	    if (e.getSource()==center) {
		center();
	    } else if (e.getSource()==restoreHidden) {
		noHide=true;
		noRecompute=false;
		sRoot=root;
		cleanTree(root);
	    }
	    else if (e.getSource()==removeRecomputed) {
		noHide=false;
		noRecompute=true;
		cleanTree(root);
	    }
	    else if (highlight==null) {
		System.err.println("Action with no highlighted goal");
		return;
	    }
	    else if (e.getSource()==hideSubtree) {
		if (highlight.property.startsWith("hidden"))
		    highlight.property=highlight.property.substring(6);
		else
		    highlight.property="hidden"+highlight.property;
	    }
	    else if (e.getSource()==showSubtree) {
		sRoot=highlight;
	    }
	    else if (e.getSource()==showCurrentGoal) {
		showCompleteGoal=true;
	    }
	    else if (e.getSource()==recomputeFast) {
		recomputeAll=false;
		jumpto=highlight;
	    }
	    else if (e.getSource()==recomputeSideEffects) { // not implemented 
		recomputeAll=true;
		jumpto=highlight;
	    }

	    repaint();
	}
	
	synchronized void cleanTree(SearchTree t) {
	    //System.err.println(t.label);
	    if (noHide & t.property.startsWith("hidden"))
		t.property=t.property.substring(6);
	    SearchTreeList l=t.children;
	    SearchTreeList p=null;
	    while (l!=null) {
		if (noRecompute & l.head.label.startsWith("recompute ")) {
		    if (p==null) t.children=l.tail;
		    else p.tail=l.tail;
		}
		else {
		    cleanTree(l.head);
		    p=l;
		}
		l=l.tail;
	    }
	}

	//////////////
	// Drawings //
	//////////////

	public synchronized void paint(Graphics g)
	{
	    if (memory==null) {
		memory = createImage(width,height);
		gMem = memory.getGraphics();
		gMem.setFont(new Font("monospaced",Font.PLAIN,12));
	    }

	    gMem.setColor(Color.white);
	    gMem.fillRect(0, 0, width, height);

	    drawAll();

	    g.drawImage(memory,0,0,this);
	}
	
	synchronized void drawAll()
	{
	    if (sRoot==null) sRoot=root;
	    if (!freezeTree.getState()) {
		fCurrent=current;
		fRoot=sRoot;
	    }
	    gMem.setColor(Color.black); //gMem.setColor(translucentGray);
	    nearString="";
	    if (fsp.getState()) {
		drawCenteredStrings=false;
		double d;
		if (fRoot!=null) d=drawSubtreeFSP(fRoot,5,-20);
	    }
	    else if (fw.getState()) {
		drawCenteredStrings=true;
		if (fRoot!=null) drawSubtreeFW(fRoot,0,-20,width);
	    }
	    else {
		drawCenteredStrings=true;
		if (fRoot!=null) drawSubtree3DTM(fRoot,0,0,-20,width,height,0);
	    }

	    if (nearString!=null) {
		gMem.setColor(Color.red);
		drawString(nearString,im,jm);
	    }
	    else {
		//highlight=null;
		//refresh=true;
		if (domains3D!=null) domains3D.paint(domains3D.getGraphics());
	    }

	    if (jumpto!=null & !recompute)
		recompute();
	}

	// Draw a tree of fixed space between leaves, returns the width used
	synchronized double drawSubtreeFSP(SearchTree t, double y, double z)
	{
	    int i=i(0,y,z);
	    int j=j(0,y,z);
	    if (i>width || j>height) return 20;

	    drawLabel(t,0,y,z);
	    if (t.property.startsWith("hidden")) return 10;

	    SearchTreeList c=t.children;
	    if (c==null) {
		if (showGoals.getState()) return t.label.length()*charSize+10;
		else return 20;
	    }
	    else  {
		c=c.reverse();
		double yc=y;
		double zc=z-20;
		double w=0;
		while (c!=null) {
		    gMem.drawLine(i,j,i(0,yc,zc),j(0,yc,zc));
		    w=w+drawSubtreeFSP(c.head,yc,zc);
		    c=c.tail;
		    yc=y+w;
		}
		if (showGoals.getState())
		    return Math.max(w,t.label.length()*charSize+10);
		else return w;
	    }
	}

	// Draw a tree in a fixed width l with equal space for each child
	synchronized void drawSubtreeFW(SearchTree t, double y, double z, double l)
	{
	    int i=i(0,y,z);
	    int j=j(0,y,z);
	    //if (i(0,y,z)<0 || i(0,y,z)>width || j(0,y,z)>height) return;
	    if (j>height) return;

	    drawLabel(t,0,y,z);
	    if (t.property.startsWith("hidden")) return;

	    //if (l<2)
	    //drawString("...", i(0,y,z-20), j(0,y,z-20));
	    //else {
		SearchTreeList c=t.children;
		if (c==null) return;
		double yc=y+l/2;
		//double zc=z-(height+z)/3;
		double zc=z-20;
		int n=c.length();
		l=l/n;
		yc=yc-l/2;
		while (c!=null) {
		    gMem.drawLine(i,j,i(0,yc,zc),j(0,yc,zc));
		    drawSubtreeFW(c.head,yc,zc,l);
		    c=c.tail;
		    yc=yc-l;
		}
		//}
	}

	// Draw a 3D tree in a fixed rectangle k l with equal space for each child
	synchronized void drawSubtree3DTM(SearchTree t, double x, double y, double z, double k, double l, int inWidth)
	{
	    int i=i(x,y,z);
	    int j=j(x,y,z);
	    if (j>height) return;

	    drawLabel(t,x,y,z);
	    if (t.property.startsWith("hidden")) return;

       	    //if (l<2)
	    //drawString("...", i(x,y,z-20), j(x,y,z-20));
	    //else {
		SearchTreeList c=t.children;
		if (c==null) return;
		double xyc,lc;
		if (inWidth<2) {
		    xyc=y+l/2;
		    lc=l;
		}
		else {
		    xyc=x+k/2;
		    lc=k;
		}
		double zc=z-20;
		int n=c.length();
		lc=lc/n;
		xyc=xyc-lc/2;
		while (c!=null) {
		    if (inWidth<2) {
			gMem.drawLine(i,j,i(x,xyc,zc),j(x,xyc,zc));
			drawSubtree3DTM(c.head,x,xyc,zc,k,lc,inWidth+1);
		    }
		    else {
			gMem.drawLine(i,j,i(xyc,y,zc),j(xyc,y,zc));
			drawSubtree3DTM(c.head,xyc,y,zc,lc,l,(inWidth+1)%4);
		    }
		    c=c.tail;
		    xyc=xyc-lc;
		}
		//}
	}

	void drawString(String s, int i, int j) {
	    if (drawCenteredStrings) {
		int n=s.length()*charSize/2;
		gMem.drawString(s,i-n,j);
	    }
	    else gMem.drawString(s,i,j);
	}

	synchronized void drawLabel(SearchTree t, double x, double y, double z) {
	    if (showCompleteGoal & highlight==t) {
		if (t==fCurrent)
		    gMem.setColor(Color.blue);
		else
		    gMem.setColor(Color.red);
		String s="";
		SearchTree ta=highlight;
		while (ta!=root) {
		    if (!ta.label.startsWith("[")) s=ta.label+","+s;
		    ta=ta.father;
		}
		drawString(s,i(x,y,z),j(x,y,z));
	    }
	    else if (t==fCurrent) {
		gMem.setColor(Color.blue);
		drawString(t.label, i(x,y,z), j(x,y,z));
	    }
	    else if (t==fRoot) {
		gMem.setColor(Color.green);
		drawString(t.label, i(x,y,z), j(x,y,z));
	    }
	    else if  (!clicked & near(i(x,y,z),j(x,y,z))) {
		nearString=nearString+t.label;
		if (highlight!=t) {
		    highlight=t;
		    refresh=true;
		    if (domains3D!=null) domains3D.paint(domains3D.getGraphics());
		}
	    }
	    else if (t==highlight || t.property.startsWith("show")) {
		gMem.setColor(Color.red);
		drawString(t.label, i(x,y,z), j(x,y,z));
	    }
	    else if (showGoals.getState()) {
		drawString(t.label, i(x,y,z), j(x,y,z));
	    }
	    else drawString("_", i(x,y,z), j(x,y,z));

	    if (t.property.startsWith("hidden")) {
		drawString("?", i(x,y,z-10), j(x,y,z-10));
	    }
	    else if (t.property.endsWith("success")) {
		gMem.setColor(Color.red);
		drawString("+", i(x,y,z-10), j(x,y,z-10));
	    }
	    gMem.setColor(Color.black);//gMem.setColor(translucentGray);
	}

	/////////////////////
	// Goals highlight //
	/////////////////////

	public synchronized void mouseMoved(MouseEvent e)
	{
	    im=e.getX();
	    jm=e.getY();
	    if (!clicked) highlight=null;
	    paint(getGraphics());
  	}
	public synchronized void mouseClicked(MouseEvent e)
	{
	    im=e.getX();
	    jm=e.getY();
	    if (highlight==null) {
		super.mouseClicked(e);		
	    }
	    else if (clicked) {
		clicked=false;
		nodes.setEnabled(false);
		showCompleteGoal=false;
	    }
	    else {
		clicked=true;
		nodes.setEnabled(true);
		showCompleteGoal=false;
	    }
	}
	public synchronized void mouseDragged(MouseEvent e)
	{
	    im=e.getX();
	    jm=e.getY();
	    super.mouseDragged(e);
	}
	public void mouseEntered(MouseEvent e) {
	    //clicked=false;
	    nodes.setEnabled(false);
	}

	boolean near(int i, int j) {
	    return ((i-im)*(i-im)+(j-jm)*(j-jm)<NEAR);
	}
    }
}

